package ru.dbotthepony.kbox2d.api

import it.unimi.dsi.fastutil.ints.IntArraySet
import ru.dbotthepony.kbox2d.collision.DynamicTree
import ru.dbotthepony.kbox2d.collision.b2_nullNode
import ru.dbotthepony.kstarbound.math.AABB

private const val DEBUG_CYCLIC_REFERENCES = true

/// A node in the dynamic tree. The client does not interact with this directly.
data class TreeNode(
	val tree: DynamicTree,
	/// Enlarged AABB
	var aabb: AABB? = null,

	var child1: Int = 0,
	var child2: Int = 0,

	// leaf = 0, free node = -1
	var height: Int = 0,

	var moved: Boolean = false,

	var userData: Any? = null,
) {
	val isLeaf get() = child1 == b2_nullNode

	// union
	private var _union: Int = b2_nullNode

	var parent by this::_union
	var next by this::_union

	internal fun validate() {
		if (!DEBUG_CYCLIC_REFERENCES)
			return

		val seen = IntArraySet()
		var parent = parent

		while (parent != b2_nullNode) {
			if (!seen.add(parent)) {
				throw IllegalStateException("Cycle detected: $seen")
			}

			seen.add(parent)
			parent = tree.nodes[parent].parent
		}
	}
}

interface IDynamicTree : IProxieable, IMovable {
	fun wasMoved(proxyID: Int): Boolean
	fun clearMoved(proxyID: Int)

	/**
	 * Validate this tree. For testing.
	 */
	fun validate()

	/**
	 * Compute the height of the binary tree in O(N) time. Should not be
	 * called often.
	 */
	val height: Int

	/// Get the maximum balance of an node in the tree. The balance is the difference
	/// in height of the two children of a node.
	val maxBalance: Int

	/// Get the ratio of the sum of the node areas to the root area.
	val getAreaRatio: Double

	/// Build an optimal tree. Very expensive. For testing.
	fun rebuildBottomUp()
}
